/* bucket.h  by craig silverstein.  started 12 februrary 1995.
 *     see bucket.cc for usage and implementation details.
 */

#ifndef BUCKET_H
#define BUCKET_H

#include "nodearc.h"         // dfn of node and arc

#ifndef ulong
typedef unsigned long ulong; // to get that extra bit
#endif

class SP;                // defined in sp.h
class Stack;             // defined in stack.h
struct Level;            // defined below; defined here for Bucket

typedef struct Bucket {
   Node *pNode;          // pointer the the list of nodes at this level
   Level *pLevel;        // what level this bucket is on
} Bucket;

typedef struct Level {
   ulong cNodes;         // number of nodes currently stored at this level
   ulong cBuckets;       // number of buckets that exist on this level
   ulong logcBuckets;    // log (base 2) of the number of buckets
   ulong logSkip;        // how many buckets a bucket at this level subtends
   ulong distExtent;     // range of distance labels that will fit on our level
   ulong minDistToBase;  // distance from our level start to top level start
   ulong maxDistToBase;  // minDistToBase + distExtent
   Bucket *rgBin;        // points to the cBuckets buckets at our level
   Bucket *pBucket;      // the current bucket at level i
} Level;

class BucketBin {
 private:
   Level *rgLevels;          // stats information about each level
   Level *topLevel;          // sentinel (last entry) for rgLevels
   double baseToZero;        // a measure of wraparound
   ulong cNodesAfterWrap;    // "Level" info for after we wrap around
   Bucket *minAfterWrap;
   long statEmpty;           // statistic: how many empty buckets we scanned
   long statExpandedNodes;   // statistic: how many nodes we expand out
   long statExpandedBuckets; // statistic: how many buckets we expand out of
   long statFirstInBucket;   // statistic: how many times we're node #1
   long statLastInBucket;    // statistic: how many times we're deleted last
   
   Bucket *DistToBucket(ulong *pDistToBase);
   Bucket *SortBucket(Level *pLevel);
   
 public:
   BucketBin(ulong minArcLen, ulong maxArcLen, ulong cLevels);
   ~BucketBin();
   void Init();              // resets bucket indices.  Assumes buckets are OK
   void dijkstra(Node *source, SP *sp);            // run dijkstra's algorithm
   void PrintStats();

   Node *Insert(Node *node, Bucket *bckNew);
   Node *Delete(Node *node, Bucket *bckOld);
   Node *RemoveMin();
   int InBucket(Node *node)        { return node->where == IN_BUCKETS; }
};

#endif
