/* constbucket.h  by craig silverstein.  started 1 november 1996.
 *     see constbucket.cc for usage and implementation details.
 */

#ifndef CONSTBUCKET_H
#define CONSTBUCKET_H

#include "nodearc.h"         // dfn of node and arc

#ifndef ulong
typedef unsigned long ulong; // to get that extra bit
#endif

#define BITS_PER_WORD   32   // you must change this for 64 bit machines!

   // We need to know the number of bits in a word, and the log of that.
   // It's the log that makes this a case statement instead of a formula.
   // MAX_LEVELS is ceil(W / LOG_W) -- assuming maxarclen < 2^W, this is
   // the maximum number of levels possible.  We allocate space for them
   // all, though we may not use it.
#if BITS_PER_WORD == 16
#define W           16
#define LOG_W       4        // I wish there was an easy formula for this!
#define MAX_LEVELS  4        // ceil(16 / 4) -- this assumes maxarclen < 2^W
#define W_MASK      15       // the bottom W bits are set (this is W-1)
#elif BITS_PER_WORD == 32
#define W           32
#define LOG_W       5
#define MAX_LEVELS  7        // ceil(32 / 5)
#define W_MASK      31
#elif BITS_PER_WORD == 64
#define W           64
#define LOG_W       6
#define MAX_LEVELS  11
#define W_MASK      63
#else
#define W           8       // This is certainly safe, though wasteful
#define LOG_W       3
#define MAX_LEVELS  3
#define W_MASK      7
#endif

class SP;                // defined in sp.h
class Stack;             // defined in stack.h

class ConstBucketBin {
 private:
   int maxLevel;                    // the levels range from 0 to maxLevel
   ulong bmOccupied[MAX_LEVELS];    // indicates non-empty buckets at each level
   int rgWhichLevel[W];             // what level we go in if we differ in bit W
   int shiftForPosInLevel[MAX_LEVELS];   // how much we shift to get level bits
   Node *rgBuckets[MAX_LEVELS][W];  // W buckets (Node *'s) allowed at each level
   int logMinArcLen;                // useful for determining bucket position
   ulong iWrapAroundAmount;         // a power of 2 > maxArcLen
   ulong iWrapAroundAmount_1;       // iWAA - 1; x & iWAA_1 == x % iWAA
   ulong cWrapArounds;              // how many times have we wrapped around?
   ulong distMinNode;               // dist of node from last call to RemoveMin
   ulong currBucket;                // bucket (at level 0) MinNode is from
   ulong currTopBucket;             // bucket (at top level) MinNode is from
   long statExpandedNodes;   // statistic: how many nodes we expand out
   long statExpandedBuckets; // statistic: how many buckets we expand out of
   long statFirstInBucket;   // statistic: how many times we're node #1
   long statLastInBucket;    // statistic: how many times we're deleted last
   
   void DistToPosition(ulong *pDistToBase, int *pLevel, int *pBucket);
   int ExpandLevels(int iLevel, int iBucket);
   
 public:
   ConstBucketBin(ulong minArcLen, ulong maxArcLen);
   void Init();              // resets bucket indices.  Assumes buckets are OK
   void dijkstra(Node *source, SP *sp);         // run Dijkstra's algorithm
   void PrintStats();

   Node *Insert(Node *node, int iLevel, int iBucket);
   Node *Delete(Node *node);
   Node *RemoveMin();
   int InBucket(Node *node)        { return node->where == IN_BUCKETS; }
};

#endif
