// translation of dikb bucket routines into c++.  Uses bucket.h headers
// link instead of bucket.cc -- user semantics are the same
// I don\'t pretend to understand everything going on here. 

#include <stdlib.h>
#include "bucket.h"

#define status temp            // all status info goes in the temp field

void Bucket::Init()
{
   rgBase[1] = 0.0;            // interior nodes first
   rgBucket[1] = rgFirst;      // current bucket
   rgmaxBucket[1] = rgFirst;   // not such a good max
   
   rgBase[0] = 0.0;            // start at the very beginning
   rgBucket[0] = rgFirst + rgcBuckets[1];
   rgmaxBucket[0] = rgFirst + rgcBuckets[1];
   minWrap = rgFirst + rgcBuckets[1] + rgcBuckets[0];
   maxWrap = rgFirst + rgcBuckets[1];

   memset(rgFirst, 0, cBuckets * sizeof(Node *));      // makes them all NULL
}

Bucket::Bucket(long maxArcLen, long cLevelsGiven)
{
   long cLogBuckets;

   rgBase = new double [2];
   rgBucket = new Node ** [2];
   rgmaxBucket = new Node ** [2];
   rgcBuckets = new long [2];
   
   cLevels = 2;             // ignore cLevelsGiven
   for ( cLogBuckets = 1; ((1 << cLogBuckets) < maxArcLen); cLogBuckets++ )
      ;
   cLogBuckets >>= 1;       // get the square root
   cLogBuckets -= 0;        // why???
   if ( cLogBuckets < 4)    // at least so many buckets
      cLogBuckets = 4;
   
   rgcBuckets[1] = 1 << cLogBuckets;       // level 1 is interior (2nd level)
   rgcBuckets[0] = ((maxArcLen+1) >> cLogBuckets) + 1;     // +1 for roundoff
   cBuckets = rgcBuckets[1] + rgcBuckets[0];

       // all the buckets are stored sequentially.  It\'s
       // <--interior buckets--><--exterior (level 0) buckets-->
   rgFirst = new Node * [cBuckets];

   Init();
}

Bucket::~Bucket()
{
   delete rgBase;
   delete rgcBuckets;
   delete rgBucket;
   delete rgmaxBucket;
   delete rgFirst;
}


Node **Bucket::ValueToBucket(Node *pNode)
{
   long pos;
   
   pos = (long)(pNode->dist - rgBase[1]);                // TRY INTERNAL LEVEL
   if ( pos >= rgcBuckets[1] )           // would go past last internal bucket
   {
      pos = (long)((pNode->dist-rgBase[0]) / rgcBuckets[1]);   // TRY EXTERNAL
      if ( pos >= rgcBuckets[0] )
	 pos -= rgcBuckets[0];           // wrap-around
      pos += rgcBuckets[1];              // get to the right part of rgFirst
   }
   else
      if ( pos < rgBucket[1] - rgFirst ) // negative cost?
	 pos = rgBucket[1] - rgFirst;    // put it in the current bucket

   return rgFirst + pos;
}


Node *Bucket::Delete(Node *node, Node **pbkOld)
{
   if ( node->next == node )    // we\'re the last one in the bucket
      *pbkOld = NULL;           // (nodes loop around to form a cycle)
   else                         // update people around us
   {
      node->prev->next = node->next;
      node->next->prev = node->prev;
      if ( *pbkOld == node )    // we might still be first
	 *pbkOld = node->next;
   }
   node->status = NOT_IN_BHEAP;
   return node;
}

Node *Bucket::Insert(Node *node, Node **pbkNew)
{
   if ( *pbkNew == NULL )          // we\'re the first one in the bucket
      *pbkNew = node->next = node->prev = node;    // a cycle of length 1
   else                            // insert us in cycle, before first
   {
      node->prev = (*pbkNew)->prev;
      node->next = *pbkNew;
      (*pbkNew)->prev->next = node;
      (*pbkNew)->prev = node;
   }
   node->status = IN_BHEAP;        // in case we weren\'t before

       // check to see if our inserted node sets new max or min records
   if ( pbkNew < rgFirst + rgcBuckets[1] )    // inserted into level 1 bucket
   {
      if ( pbkNew > rgmaxBucket[1] )	rgmaxBucket[1] = pbkNew; 
   }
   else if ( pbkNew >= rgBucket[0] )          // level 0 bucket, no wraparound
   {
      if ( pbkNew > rgmaxBucket[0] )	rgmaxBucket[0] = pbkNew;
   }
   else                                       // level 0 bucket, wrapped around
   {
      if ( pbkNew > maxWrap )        	maxWrap = pbkNew;
      if ( pbkNew < minWrap )        	minWrap = pbkNew;
   }
      
   return node;
}

Node *Bucket::MoveUp(Node *node, Node **pbkOld, Node **pbkNew)
{
   if ( pbkOld )                 // we can insert without removing
      Delete(node, pbkOld);
   if ( pbkNew )                 // and vice-versa
      Insert(node, pbkNew);
   return node;
}

Node **Bucket::SortBucket(long iLevel)
// pFirstLevel is the first node of level 0 (external buckets) in rgFirst
{
   Node *node, *nextNode;           // used to traverse the bucket for sorting
   Node **bkNew;
   Node **minBucket;                // for internal level
   Node **pFirstLevel = rgFirst + rgcBuckets[1];

   rgBase[1] = rgBase[0] + (rgBucket[0] - pFirstLevel) * rgcBuckets[1];
   rgmaxBucket[1] = rgFirst;
   rgBucket[1] = rgFirst;                    // the min to start with
   minBucket = rgFirst + rgcBuckets[1];      // overestimate of the minimum
   
   node = *(rgBucket[0]);           // the bucket we\'re sorting
   *(rgBucket[0]) = NULL;           // remember to do this now
   node->prev->next = NULL;         // turn our ring into a list
   for ( ; node; node = nextNode )
   {
      nextNode = node->next;        // remember to do this now too
      bkNew = ValueToBucket(node);
      if ( bkNew < minBucket )      // set the minimum
	 minBucket = bkNew;
      Insert(node, bkNew);
   }
   rgBucket[1] = minBucket;         // we start with the minimum
   return rgBucket[1];              // the minimum bucket that has something
}

Node *Bucket::RemoveMin()
{
   Node **retBucket;       // lowest bucket holding a node
   int iWrapped;           // 1 after we wrap around at the top level once

         // first we try to find an internal bucket with a node in it
   for ( ; rgBucket[1] <= rgmaxBucket[1]; rgBucket[1]++ )
      if ( *(rgBucket[1]) != NULL )
	 return Delete( *(rgBucket[1]), rgBucket[1] );  // remove and return
   
         // no luck: try to find an external bucket
   for ( iWrapped = 0; iWrapped <= 1; iWrapped++ )      // wrap once
   {
      for ( ; rgBucket[0] <= rgmaxBucket[0]; rgBucket[0]++ )
	 if ( *(rgBucket[0]) != NULL )
	 {
	    retBucket = SortBucket(0);
	    rgBucket[0]++;                              // done with bucket
	    return Delete( *retBucket, retBucket );
	 }

         // no luck: try wrapping around on the external buckets
      rgBucket[0] = minWrap;
      rgmaxBucket[0] = maxWrap;
      minWrap = rgFirst + rgcBuckets[1] + rgcBuckets[0];
      maxWrap = rgFirst + rgcBuckets[1];
      rgBase[0] += rgcBuckets[0] * rgcBuckets[1];
   }

   return NULL;                                         // no node found
}
