/* heap.h  by craig silverstein.  started 10 february 1995.
 *     see heap.cc for description.
 */

#ifndef HEAP_H
#define HEAP_H

#include "nodearc.h"     // dfn of node and arc

class SP;                // defined in sp.h
class Stack;             // defined in stack.h

#define NODE_POS(pNode) ( (pNode)->info.heapInfo.nodePos )

class Heap {
 private:
   long size;             // how many nodes are currently in the heap
   long maxSize;          // how many nodes are allowed (set in Init )
   Node **data;           // array of pointers to nodes
   unsigned long statOps; // number of heap operations: MoveUp, Insert
   unsigned long statSwaps;   // number of swaps: the primitive heap op
   
 public:
   Heap(long maxSizeGiven);
   ~Heap();
   void dijkstra(Node *source, SP *sp);// run dijkstra's algo
   void MoveUp(Node *node);            // reassigns node w/ new value and sifts
   void MoveDown(Node *node);          // reassigns node w/ new value and sifts
   void Heapify();                     // puts all nodes into heap order
   void Insert(Node *node);            // adds--TO END--of heap
   Node *FindMin()                     // returns top element
        { return data[0]; }
   Node *RemoveArbitrary();            // returns an arbitrary element
   Node *RemoveMin();                  // returns top element and sifts
   void Print();                       // debugging: print heap contents
   int IsEmpty()     
         { return size == 0; }         // nothing in our heap
   int IsFull()
         { return size == maxSize; }   // can't fit another thing in our heap
   void Empty()  { statOps = 0; size = 0; }         // empty our heap
   int InHeap(Node *node)
         { return node->where == IN_HEAP; }
   void PrintStats();                  // no special stats to print
};

#endif
