/* nodearc.h  by craig silverstein.  Started 14 february 1995.
 *     Contains the definition of the node and arc class needed
 *     for all the code.
 *     To make code more comprehensible, we use a union for data
 *     types limited to certain algorithms.  We include a #define
 *     for data types limited to certain problems (min-cost flow, etc).
 */

#ifndef NODEARC_H
#define NODEARC_H

#define IN_NONE       0   // possible data structures the node can be in
#define IN_HEAP       1
#define IN_GOR        2
#define IN_BUCKETS    4

typedef struct Node;      // so the arc knows about it
   
typedef struct Arc {
   double len;            // length of arc -- double avoids precision trouble
   struct Node *head;     // where the arc ends up
#if defined(MAXFLOW) || defined(MINCOST)
   long cap;              // capacity
   long r_cap;            // residual capacity
   struct Arc *sister;    // opposite arc in residual graph
#endif
} Arc;

typedef struct Node {
   double dist;           // tentative shortest path length to some node
   Arc   *first;          // first outgoing arc
   Arc   *current;        // used to keep place of arcs during recursion
   struct Node *parent;   // parent on the heap
   unsigned long where;   // what data structure we're in:  IN_* (above)
   union {
      struct {
	 struct Node *next;          // next in bucket
	 struct Node *prev;          // prev in bucket
	 void *bucket;               // you should cast this
	 unsigned long distToBase;   // dist since last wraparound
      } bckInfo;                     // FOR BUCKETS
      struct {
	 struct Node *next;          // next in bucket
	 struct Node *prev;          // prev in bucket
	 int iLevel;                 // what level the node is in
         int iBucket;                // the bucket, in the level, the node is in
	 unsigned long distToBase;   // dist since last wraparound
      } constbckInfo;                // FOR CONST_BUCKETS
      struct {
	 unsigned long nodePos;      // where it is in the heap
	 long lvlBucket;             // where it would be if it was in a bucket
      } heapInfo;                    // FOR HEAPS and HOT-QUEUES
      long status;                   // FOR GOR
   } info;                           // do you have a better name?
#ifdef MINCOST
   long excess;           // excess flow currently held in a node
#endif
} Node;

#endif
