#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>     
#include <sys/resource.h>
#include <unistd.h>

/*********************************************************************/
/*                                                                   */
/* SETS memory limit so the process will die rather than page.       */
/* This makes sure timing values aren't off because of paging,       */
/* and also that the experiments don't take too long to run.         */
/* needs: <sys/time.h>, <sys/resource.h>, <unistd.h>                 */
/* depends strongly on compiler and OS                               */
/*    for linux we use the proc/meminfo file, which looks like this: */
/*         total:    used:    free:  shared: buffers:  cached:       */
/* Mem:  31756288 30646272  1110016 11972608  2224128 10637312       */
/* Swap: 105902080 28020736 77881344                                 */
/*                                                                   */
/*********************************************************************/

#define MEM_USE    .7    /* we allow this much use of main memory */

void setlimit()
{
#ifdef __linux
   struct rlimit r;
   FILE *fpMem;          /* part of linux's PROC file system */
   char szLine[255];
   unsigned long cchMem;

   fpMem = fopen("/proc/meminfo", "r");
   if ( !fpMem )              /* can't figure out how much memory there is */
      return;
   fgets(szLine, 255, fpMem);                         /* ignore first line */
   fgets(szLine, 255, fpMem);                          /* read second line */
   cchMem = atoi(szLine + strlen("Mem:"));         /* skip past "Mem:" bit */
   fclose(fpMem);
   if ( !cchMem )                /* we don't know how much memory there is */
      return;
   r.rlim_max = r.rlim_cur = (int)(cchMem * MEM_USE);
   setrlimit(RLIMIT_DATA, &r);                 /* set size of data segment */
#endif
}
