// sp.h  by craig silverstein.  Started 23 february 1995.
//     The main driver for shortest-path routines.  Whenever you want
//     to use an SP, you include sp.h and call sp::sp with the
//     algorithm you want, taken from the #defines below.

#ifndef SP_H
#define SP_H

#include <stdlib.h>
#include <limits.h>         // where VERY_FAR is defined
#include "nodearc.h"
#include "stack.h"
#include "bucket.h"
#include "hotqueue.h"
#include "heap.h"
#include "constbucket.h"

#define VERY_FAR            1e+300     // hopefully big enough; hopefully a dbl

#define SP_DIK_HEAP         1
#define SP_GOR              2
#define SP_DIK_CONSTBUCKETS 4
#define SP_DIK_BUCKETS      256        // use SP_D_B + # of buckets levels
#define SP_DIK_HOTQUEUE     512        // use SP_H_Q + # of buckets levels
//** add new sp algorithm type-number here **//
              // example for buckets: SP(n, nodes, SP_DIK_BUCKETS + 2)

#define DIK_BUCKETS_DEFAULT  2         // default of two bucket levels


int sp_openarc(Arc *arc);              // whether SP thinks an arc is usable
extern double ArcLen(long cNodes, Node *nodes,
		     double *pMin = NULL, double *pMax = NULL);


class SP {
 private:
   long cNodes;
   Node *nodes;
   int spType;
   void Init(Node *source, Stack *stkSources, int tmp);

   BucketBin *bucket;                 // for dijkstra_buckets
   Heap *heap;                        // for dijkstra_heap
   Stack *stkSources, *stkTopSort;    // for GOR
   HOTQueue *hotqueue;                // for dijkstra_hotqueue
   ConstBucketBin *constbucket;       // for dijkstra_constbucket
//** add new SP data structure here **//

   void gor(Node *source);

 public:
   SP(long cNodesGiven, Node *nodesGiven, int spTypeGiven);
   ~SP();
   void sp(Node *source);
      // these must be public, alas, so Heap and Bucket can modify them
   long cCalls;         // # of times SP has been called since initialization
   long cScans;         // # of nodes SP algorithm has looked at (since init)
   long cUpdates;       // # of times a node value was lowered (since init)
   void PrintStats();
};

#endif
